@@ -116,6 +116,7 @@ if (MIME::Charset::USE_ENCODE) {
 	}
     }
 } else {
+    require Unicode::String;
     require MIME::Charset::_Compat;
     for my $sub (@ENCODE_SUBS) {
         no strict "refs";
@@ -130,7 +131,7 @@ if (MIME::Charset::USE_ENCODE) {
 #------------------------------
 
 ### The package version, both in 1.23 style *and* usable by MakeMaker:
-$VERSION = '1.012.3';
+$VERSION = '1.015.0';
 
 ### Public Configuration Attributes
 $Config = {
@@ -142,7 +143,11 @@ $Config = {
     MaxLineLen => 76,
     Minimal => 'YES',
 };
-eval { require MIME::EncWords::Defaults; };
+eval {
+    local @INC = @INC;
+    pop @INC if $INC[-1] eq '.';
+    require MIME::EncWords::Defaults;
+};
 
 ### Private Constants
 
@@ -152,7 +157,37 @@ my $NONPRINT = qr{[^$PRINTABLE]}; # Improvement: Unicode support.
 my $UNSAFE = qr{[^\x01-\x20$PRINTABLE]};
 my $WIDECHAR = qr{[^\x00-\xFF]};
 my $ASCIITRANS = qr{^(?:HZ-GB-2312|UTF-7)$}i;
-my $DISPNAMESPECIAL = "\\x22(),:;<>\\x40\\x5C"; # RFC5322 name-addr specials.
+my $ASCIIINCOMPAT = qr{^UTF-(?:16|32)(?:BE|LE)?$}i;
+my $DISPNAMESPECIAL = qr{[()<>\[\]:;@\\,."]}; # RFC5322 name-addr specials.
+
+#------------------------------
+
+# _utf_to_unicode CSETOBJ, STR
+#     Private: Convert UTF-16*/32* to Unicode or UTF-8.
+sub _utf_to_unicode {
+    my $csetobj = shift;
+    my $str = shift;
+
+    return $str if is_utf8($str);
+
+    return $csetobj->decode($str)
+	if MIME::Charset::USE_ENCODE();
+
+    my $cset = $csetobj->as_string;
+    my $unistr = Unicode::String->new();
+    if ($cset eq 'UTF-16' or $cset eq 'UTF-16BE') {
+	$unistr->utf16($str);
+    } elsif ($cset eq 'UTF-16LE') {
+	$unistr->utf16le($str);
+    } elsif ($cset eq 'UTF-32' or $cset eq 'UTF-32BE') {
+	$unistr->utf32($str);
+    } elsif ($cset eq 'UTF-32LE') {
+	$unistr->utf32le($str);
+    } else {
+	croak "unknown transformation '$cset'";
+    }
+    return $unistr->utf8;
+}
 
 #------------------------------
 
@@ -225,7 +260,7 @@ However, adjacent encoded-words with same charset will be concatenated
 to handle multibyte sequences safely.
 
 B<**>
-Language information defined by RFC2231, section 5 will be additonal
+Language information defined by RFC2231, section 5 will be additional
 third element, if any.
 
 B<*>
@@ -714,11 +749,18 @@ sub encode_mimewords  {
     # unsafe ASCII sequences
     my $UNSAFEASCII = ($maxrestlen <= 1)?
 	qr{(?: =\? )}ox:
-	qr{(?: =\? | [$PRINTABLE]{$Params{MaxLineLen}} )}ox;
-    $UNSAFEASCII = qr{(?: [$DISPNAMESPECIAL] | $UNSAFEASCII )}ox
+	qr{(?: =\? | [$PRINTABLE]{$Params{MaxLineLen}} )}x;
+    $UNSAFEASCII = qr{(?: $DISPNAMESPECIAL | $UNSAFEASCII )}x
 	if $Params{Minimal} eq 'DISPNAME';
 
     unless (ref($words) eq "ARRAY") {
+	# workaround for UTF-16* & UTF-32*: force UTF-8.
+	if ($charsetobj->as_string
+	    and $charsetobj->as_string =~ /$ASCIIINCOMPAT/) {
+	    $words = _utf_to_unicode($charsetobj, $words);
+	    $charsetobj = MIME::Charset->new('UTF-8');
+	}
+
 	my @words = ();
 	# unfolding: normalize linear-white-spaces and orphan newlines.
 	$words =~ s/(?:[\r\n]+[\t ])*[\r\n]+([\t ]|\Z)/$1? " ": ""/eg;
@@ -735,7 +777,8 @@ sub encode_mimewords  {
 
 		# workaround for ``ASCII transformation'' charsets
 		my $u = $w;
-		if ($charsetobj->as_string =~ /$ASCIITRANS/) {
+		if ($charsetobj->as_string
+		    and $charsetobj->as_string =~ /$ASCIITRANS/) {
 		    if (MIME::Charset::USE_ENCODE) {
 			if (is_utf8($w) or $w =~ /$WIDECHAR/) {
 			    $w = $charsetobj->undecode($u);
@@ -786,6 +829,13 @@ sub encode_mimewords  {
 	next unless length($s);
 	my $csetobj = MIME::Charset->new($cset || "",
 					 Mapping => $Params{Mapping});
+
+	# workaround for UTF-16*/UTF-32*: force UTF-8
+	if ($csetobj->as_string and $csetobj->as_string =~ /$ASCIIINCOMPAT/) {
+	    $s = _utf_to_unicode($csetobj, $s);
+	    $csetobj = MIME::Charset->new('UTF-8');
+	}
+
 	# determine charset and encoding
 	# try defaults only if 7-bit charset detection is not required
 	my $enc;
@@ -873,14 +923,14 @@ sub encode_mimewords  {
 	} elsif ($Params{Encoding} eq 'A' and ! $b and $q) {
 	    $enc = 'Q';
 	# Otherwise, assuming 'Q', when characters to be encoded are more than
-	# 6th of total, 'B' will win.
+	# 6th of total (plus a little fraction), 'B' will win.
 	# Note: This might give 'Q' so great advantage...
 	} else {
 	    my @no_enc = grep { ! $_->[1] } @triplets;
 	    my $total = length join('', map { $_->[0] } (@no_enc, @s_enc));
 	    my $q = scalar(() = join('', map { $_->[0] } @s_enc) =~
 			   m{[^- !*+/0-9A-Za-z]}g);
-	    if ($total < $q * 6) {
+	    if ($total + 8 < $q * 6) {
 		$enc = 'B';
 	    } else {
 		$enc = 'Q';
@@ -897,24 +947,26 @@ sub encode_mimewords  {
     }
 
     # Split long ``words''.
-    my @splitted;
+    my @splitwords;
     my $restlen;
     if ($Params{MaxLineLen} < 0) {
-      @splitted = @triplets;
+      @splitwords = @triplets;
     } else {
       $restlen = $firstlinelen;
       foreach (@triplets) {
 	my ($s, $enc, $csetobj) = @$_;
 
-	push @splitted, &_split($s, $enc, $csetobj, $restlen, $maxrestlen);
-	my ($last, $lastenc, $lastcsetobj) = @{$splitted[-1]};
+	my @s = &_split($s, $enc, $csetobj, $restlen, $maxrestlen);
+	push @splitwords, @s;
+	my ($last, $lastenc, $lastcsetobj) = @{$s[-1]};
 	my $lastlen;
 	if ($lastenc) {
 	    $lastlen = $lastcsetobj->encoded_header_len($last, $lastenc);
 	} else {
 	    $lastlen = length($last);
 	}
-	$restlen -= $lastlen; # FIXME: Sometimes estimated longer
+	$restlen = $maxrestlen if scalar @s > 1; # has split; new line(s) fed
+	$restlen -= $lastlen;
 	$restlen = $maxrestlen if $restlen <= 1;
       }
     }
@@ -922,7 +974,7 @@ sub encode_mimewords  {
     # Do encoding.
     my @lines;
     $restlen = $firstlinelen;
-    foreach (@splitted) {
+    foreach (@splitwords) {
 	my ($str, $encoding, $charsetobj) = @$_;
 	next unless length($str);
 
@@ -978,10 +1030,10 @@ sub _split {
 	return ([$str, $encoding, $charset]);
     }
 
-    my (@splitted, $ustr, $first);
+    my (@splitwords, $ustr, $first);
     while (length($str)) {
 	if ($charset->encoded_header_len($str, $encoding) <= $restlen) {
-	    push @splitted, [$str, $encoding, $charset];
+	    push @splitwords, [$str, $encoding, $charset];
 	    last;
 	}
 	$ustr = $str;
@@ -996,10 +1048,10 @@ sub _split {
 	    ($first, $str) = &_clip_unsafe($ustr, $encoding, $charset,
 					   $maxrestlen);
 	}
-	push @splitted, [$first, $encoding, $charset];
+	push @splitwords, [$first, $encoding, $charset];
 	$restlen = $maxrestlen;
     }
-    return @splitted;
+    return @splitwords;
 }
 
 # _split_ascii RAW, ROOM_OF_FIRST_LINE, MAXRESTLEN
@@ -1012,12 +1064,13 @@ sub _split_ascii {
     my $maxrestlen = shift;
     $restlen ||= $maxrestlen;
 
-    my @splitted;
+    my @splitwords;
     my $ascii = MIME::Charset->new("US-ASCII", Mapping => 'STANDARD');
     foreach my $line (split(/(?:[\t ]*[\r\n]+)+/, $s)) {
         my $spc = '';
 	foreach my $word (split(/([\t ]+)/, $line)) {
-	    next unless scalar(@splitted) or $word; # skip first garbage
+	    # skip first garbage
+	    next unless scalar(@splitwords) or defined $word;
 	    if ($word =~ /[\t ]/) {
 		$spc = $word;
 		next;
@@ -1026,33 +1079,33 @@ sub _split_ascii {
 	    my $cont = $spc.$word;
 	    my $elen = length($cont);
 	    next unless $elen;
-	    if (scalar(@splitted)) {
+	    if (scalar(@splitwords)) {
 		# Concatenate adjacent words so that encoded-word and
 		# unencoded text will adjoin with separating whitespace.
 		if ($elen <= $restlen) {
-		    $splitted[-1]->[0] .= $cont;
+		    $splitwords[-1]->[0] .= $cont;
 		    $restlen -= $elen;
 		} else {
-		    push @splitted, [$cont, undef, $ascii];
+		    push @splitwords, [$cont, undef, $ascii];
 		    $restlen = $maxrestlen - $elen;
 		}
 	    } else {
-		push @splitted, [$cont, undef, $ascii];
+		push @splitwords, [$cont, undef, $ascii];
 		$restlen -= $elen;
 	    }
 	    $spc = '';
 	}
 	if ($spc) {
-	    if (scalar(@splitted)) {
-		$splitted[-1]->[0] .= $spc;
+	    if (scalar(@splitwords)) {
+		$splitwords[-1]->[0] .= $spc;
 		$restlen -= length($spc);
 	    } else { # only WSPs
-		push @splitted, [$spc, undef, $ascii];
+		push @splitwords, [$spc, undef, $ascii];
 		$restlen = $maxrestlen - length($spc);
 	    }
 	}
     }
-    return @splitted;
+    return @splitwords;
 }
 
 # _clip_unsafe UNICODE, ENCODING, CHARSET_OBJECT, ROOM_OF_FIRST_LINE
@@ -1070,7 +1123,7 @@ sub _clip_unsafe {
     # Seek maximal division point.
     my ($shorter, $longer) = (0, length($ustr));
     while ($shorter < $longer) {
-	my $cur = int(($shorter + $longer + 1) / 2);
+	my $cur = ($shorter + $longer + 1) >> 1;
 	my $enc = substr($ustr, 0, $cur);
 	if (MIME::Charset::USE_ENCODE ne '') {
 	    $enc = $charset->undecode($enc);
@@ -1155,7 +1208,6 @@ sub _getparams {
     return %GotParams;
 }
 
-
 #------------------------------
 
 =back
